/* src/stack.c */
/*
 * OCSP responder
 * by Massimiliano Pala (madwolf@openca.org)
 * OpenCA Licensed Software
 *
 * Copyright (c) 2001-2006 The OpenCA Project.  All rights reserved.
 *
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include "stack.h"
#include <stdlib.h>

static STACK_NODE * _STACK_NODE_new( void *data ) {
	STACK_NODE *ret = NULL;

	if((ret = (STACK_NODE *) calloc (1, sizeof(STACK))) == NULL) {
		return(NULL);
	};

	ret->next = NULL;
	ret->prev = NULL;
	ret->data = data;

	return (ret);
}

static int _STACK_NODE_free ( STACK_NODE *n ) {
	if( !n ) return (STACK_ERR);

	free( n );

	return(0);
}

STACK * STACK_new( void ) {
	STACK *ret = NULL;

	if((ret = (STACK *) calloc (1, sizeof(STACK))) == NULL) {
		return(NULL);
	};

	ret->head = NULL;
	ret->tail = NULL;
	ret->elements = 0;

	return(ret);
};

int STACK_free ( STACK * st ) {

	STACK_NODE *n = NULL;
	STACK_NODE *nn = NULL;

	if( !st ) {
		return(STACK_ERR);
	}

	n = st->head;
	while( n != NULL ) {
		if( n->data ) free (n->data);
		if( n ) nn = n->next;

		_STACK_NODE_free( n );
		n = nn;
	}

	free ( st );

	return(0);
};

void * STACK_pop ( STACK *st ) {

	STACK_NODE *n = NULL;
	void *data = NULL;

	if( (st == NULL) || ( st->tail == NULL) )
					return(NULL);

	n = st->tail;
	st->tail = n->prev;

	data = n->data;

	if( st->tail == NULL ) {
		st->head = NULL;
		st->elements = 0;
	}

	if( _STACK_NODE_free ( n ) == 0 ) {
		st->elements--;
	}

	return(data);
};

int STACK_push ( STACK *st, void *obj ) {
	STACK_NODE *n = NULL;

	if( (st == NULL) || ( obj == NULL) ) return(STACK_ERR);

	if((n = _STACK_NODE_new( obj )) == NULL ) {
		return(STACK_ERR);
	}

	n->data = obj;

	if( st->tail ) {
		n->prev = st->tail;
		n->next = NULL;

		st->tail->next = n;
		st->tail = n;
	} else {
		st->tail = n;
		st->head = n;
	}

	st->elements++;

	return(st->elements);
};

int STACK_elements ( STACK *st ) {
	if( st == NULL ) return (STACK_ERR);

	return(st->elements);
};

void * STACK_get_num ( STACK *st, int num ) {
	STACK_NODE *n = NULL;
	int i;

	if( (st == NULL) || (st->elements < num)) return(NULL);

	i = 0; n = st->head;
	while( n ) {
		if( i == num ) return( n->data );
		n=n->next;
		i++;
	}

	return(NULL);
};

int STACK_ins_num ( STACK *st, int num, void *obj ) {
	STACK_NODE *n = NULL;
	STACK_NODE *new_n = NULL;

	int i;

	if( (st == NULL) || (num > st->elements) || (obj == NULL ))
		return(STACK_ERR);

	n = st->head; i = 0;
	while( n ) {
		if( i == num ) break;
		if( i > num ) return (STACK_ERR);
		n = n->next;
		i++;
	}

	if((new_n = _STACK_NODE_new ( obj )) == NULL ) {
		return(STACK_ERR);
	}

	new_n->next = n;
	new_n->prev = n->prev;

	n->prev = new_n;

	if( num == 0 ) {
		st->head = new_n;
	} else {
		new_n->prev->next = new_n;
	}

	st->elements++;
	
	return(STACK_OK);
};

void * STACK_pop_num ( STACK *st, int num ) {
	STACK_NODE *tmp_n = NULL;
	STACK_NODE *n = NULL;

	int i;
	void *obj = NULL;

	if( st == NULL ) return (NULL);

	tmp_n = st->head; i = 0;
	while ( tmp_n ) {
		if ( i == num ) {
			n = tmp_n;
			break;
		}
		i++;
		tmp_n = tmp_n->next;
	}
	if( n == NULL) return( NULL);

	if( n->prev != NULL ) {
		n->prev->next = n->next;
	}

	if( n->next ) {
		n->next->prev = n->prev;
	}

	if( num == 0 ) {
		st->head = n->next;
	}

	if( num == STACK_elements( st ) ) {
		st->tail = n->prev;
	}

	obj = n->data;

	_STACK_NODE_free( n );
	st->elements--;
	
	return(obj);
};


